/******************************************************************************/
/*                                                                            */
/*  Cedocida DV-Codec                                                         */
/*                                                                            */
/*  Copyright (C) 2005 Andreas Dittrich <dittrich@eit.uni-kl.de>              */
/*  Project start: 15.12.2003                                                 */
/*  Last change  : 15.06.2007                                                 */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  This program is free software; you can redistribute it and/or modify      */
/*  it under the terms of the GNU General Public License as published by      */
/*  the Free Software Foundation; either version 2 of the License, or         */
/*  (at your option) any later version.                                       */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*                                                                            */
/******************************************************************************/

#include <stdlib.h>
#include <memory.h>

#include "inttypes.h"

#include "signalproc.h"
#include "dv_spec.h"
#include "dct/dct.h"
#include "vlc/decode_vlc.h"
#include "decode_tables.h"
#include "config.h"
#include "decode.h"

#include "convert/convert.h"


#define LIMIT_1_TO_254( x ) ( ((x)<1) ? 1 : ( ((x)>254) ? 254 : (x) ) )
#define LIMIT_2_TO_508( x ) ( ((x)<2) ? 2 : ( ((x)>508) ? 508 : (x) ) )
#define LIMIT_MINUS255_TO_PLUS255( x ) ( ((x)<-255) ? (-255) : ( ((x)>255) ? (255) : (x) ) )


typedef struct {
	uint8_t *buffer_base;
	int16_t *buffer_5_MBs;
	uint32_t *buffer_frame;

	int color_format;
	int flip_RGB;
	int is_PAL;

	configvars_t lpmyconfig[1];
} decoder_context_t;


static void decode_PAL_DVSD( void *hnd, uint8_t *src_DV, uint32_t *dst );
static void decode_NTSC_DVSD( void *hnd, uint8_t *src_DV, uint32_t *dst );

static void decode_DIF_videosegment( uint8_t *DIF_block, int16_t *dst_5_MBs, 
								 uint32_t DIF_sequence_number, uint32_t DIF_block_number );


static void deinterleave_MB_YV12_PAL(uint32_t *src, uint32_t *dst_YUV420_Y, int dst_YUV420_Y_pitch, 
						  uint32_t *dst_YUV420_U, int dst_YUV420_U_pitch, 
					      uint32_t *dst_YUV420_V, int dst_YUV420_V_pitch
					      );

static void deinterleave_MB_YV12_NTSC(uint32_t *src, uint32_t *dst_YUV420_Y, int dst_YUV420_Y_pitch, 
						  uint32_t *dst_YUV420_U, int dst_YUV420_U_pitch, 
					      uint32_t *dst_YUV420_V, int dst_YUV420_V_pitch
					      );

static void deinterleave_MBborder_YV12_NTSC(uint32_t *src, uint32_t *dst_YUV420_Y, int dst_YUV420_Y_pitch, 
						  uint32_t *dst_YUV420_U, int dst_YUV420_U_pitch, 
					      uint32_t *dst_YUV420_V, int dst_YUV420_V_pitch
					      );

static void YUV420_to_YUY2( uint8_t* src_YUV420_Y, int src_YUV420_Y_pitch, 
					 uint8_t* src_YUV420_U, int src_YUV420_U_pitch, 
					 uint8_t* src_YUV420_V, int src_YUV420_V_pitch,
					 uint8_t* dst_YUY2, int dst_YUY2_pitch, int width, int height );

static void YUV411_to_YUY2( uint8_t* src_YUV420_Y, int src_YUV420_Y_pitch, 
					 uint8_t* src_YUV420_U, int src_YUV420_U_pitch, 
					 uint8_t* src_YUV420_V, int src_YUV420_V_pitch,
					 uint8_t* dst_YUY2, int dst_YUY2_pitch, int width, int height );







void * decoder_open()
{
	decoder_context_t *pCtx;

	pCtx = (decoder_context_t*)malloc( sizeof(decoder_context_t) );
	if (pCtx==NULL)
		return NULL;

#define LOG2_ALIGN 4
#define SIZE_BUFFER_5_MBS (5*6*64*2)
#define SIZE_BUFFER_FRAME (720*576*4)

	pCtx->buffer_base = (uint8_t*)malloc( ((1<<LOG2_ALIGN)-1) + SIZE_BUFFER_FRAME + SIZE_BUFFER_5_MBS );
	if (pCtx->buffer_base==NULL)
		decoder_close( pCtx );

	pCtx->buffer_frame = (uint32_t*)(  ( (int64_t)(pCtx->buffer_base) + ((1<<LOG2_ALIGN)-1) ) & (~((1<<LOG2_ALIGN)-1))  );
	pCtx->buffer_5_MBs = (int16_t*)( (uint8_t*)(pCtx->buffer_frame) + SIZE_BUFFER_FRAME );

	pCtx->color_format = COLORFORMAT_YV12;
	pCtx->flip_RGB = 0;
	pCtx->is_PAL = 0;

	return pCtx;
}


void decoder_configure( void *hnd, int color_format, int flip_RGB, int is_PAL, configvars_t *lpmyconfig )
{
	decoder_context_t *pCtx = (decoder_context_t*)hnd;

	if (pCtx==NULL)
		return;

	pCtx->color_format = color_format;
	pCtx->flip_RGB = flip_RGB;
	pCtx->is_PAL = is_PAL;

	memcpy( pCtx->lpmyconfig, lpmyconfig, sizeof(configvars_t) );
}


void decoder_close( void *hnd )
{
	decoder_context_t *pCtx = (decoder_context_t*)hnd;

	if (pCtx==NULL)
		return;

	if (pCtx->buffer_base!=NULL)
		free(pCtx->buffer_base);

	free(pCtx);
	return;
}


void decode_DVSD( void *hnd, uint8_t *src_DV, uint32_t *dst )
{
	decoder_context_t *pCtx = (decoder_context_t*)hnd;

	if (pCtx==NULL)
		return;

	if (pCtx->is_PAL)
		decode_PAL_DVSD( pCtx, src_DV, dst );
	else
		decode_NTSC_DVSD( pCtx, src_DV, dst );

	emms_asm();
}


static void decode_PAL_DVSD( void *hnd, uint8_t *src_DV, uint32_t *dst )
{
	decoder_context_t *pCtx = (decoder_context_t*)hnd;

	int color_format = pCtx->color_format;

	int i, r, j, n;
	int16_t *dst_5_MBs = pCtx->buffer_5_MBs;
	uint32_t *dst_step1 = dst;
	uint32_t *dst_step2 = pCtx->buffer_frame;

	if (color_format==COLORFORMAT_YUY2)
	{
		dst_step1 = pCtx->buffer_frame;
	}

	for (i=0; i<12; i++)	// sequence
	{
		for (r=0; r<3; r++)	// row_MB in SB
		{
			for (j=0; j<9; j++)	// col_MB in SB
			{
				int block_indx = 5*SB_row_order[r][j]; 
				int DBN_1st = (7+block_indx+block_indx/15);	// first DIF_blocknumber
				uint8_t *src_DIF_videosegment = src_DV + i*(150*80) + DBN_1st*(80);

				decode_DIF_videosegment( src_DIF_videosegment, dst_5_MBs, (uint8_t)i, DBN_1st );

				for (n=0; n<5; n++)		// col_SB
				{
					// save it first as YV12 and then convert to other formats
					// todo: direct conversion to the right format
					uint32_t *src = (uint32_t*)( dst_5_MBs + n*6*64 );

					uint32_t *dst_Y = dst_step1; int dst_Y_pitch = (720/4);
					uint32_t *dst_V = dst_Y + 576*dst_Y_pitch; int dst_V_pitch = (360/4);
					uint32_t *dst_U = dst_V + 288*dst_V_pitch; int dst_U_pitch = dst_V_pitch;

					static const int modulo_offset[5] = { 2, 6, 8, 0, 4};
					static const int j_SB[5] = { 2, 1, 3, 0, 4};

					int row_SB = (i+modulo_offset[n])%12;

					uint32_t *dst_YUV420_Y= dst_Y  + (j_SB[n])*(9*16/4) + j*(16/4) + (row_SB*3+r)*16*dst_Y_pitch;
					uint32_t *dst_YUV420_V = dst_V + (j_SB[n])*(9*8/4)  + j*(8/4) + (row_SB*3+r)*8*dst_V_pitch;
					uint32_t *dst_YUV420_U = dst_U + (j_SB[n])*(9*8/4)  + j*(8/4) + (row_SB*3+r)*8*dst_U_pitch;


					deinterleave_MB_YV12_PAL(src, dst_YUV420_Y, dst_Y_pitch, 
						dst_YUV420_U, dst_U_pitch, dst_YUV420_V, dst_V_pitch );

				}
			}
		}
	}

	if (color_format==COLORFORMAT_YV12)
	{
		if (pCtx->lpmyconfig->decoder_YV12_chroma_sampling!=YV12_CHROMASAMPLING_DV)
		{
			YUV420_to_YUY2( (uint8_t*)dst_step1, 720, 
				(uint8_t*)(dst_step1+576*720/4+288*90), 360, 
				(uint8_t*)(dst_step1+576*720/4), 360,
				(uint8_t*)dst_step2, 720*2, 
				720, 576 );

			if (pCtx->lpmyconfig->decoder_YV12_chroma_sampling==YV12_CHROMASAMPLING_MPEG2_I)
			{
				mmx_yuy2_i_to_yv12( (uint8_t*)dst_step2, 720*2, 720*2, 
					(uint8_t*)dst, (uint8_t*)dst+576*720+288*360, (uint8_t*)dst+576*720, 720, 360,
					576 );
			}
			else
			{
				mmx_yuy2_to_yv12( (uint8_t*)dst_step2, 720*2, 720*2, 
					(uint8_t*)dst, (uint8_t*)dst+576*720+288*360, (uint8_t*)dst+576*720, 720, 360,
					576 );
			}
		}
	}
	else if (color_format==COLORFORMAT_YUY2)
	{
		YUV420_to_YUY2( (uint8_t*)dst_step1, 720, 
			(uint8_t*)(dst_step1+576*720/4+288*90), 360, 
			(uint8_t*)(dst_step1+576*720/4), 360,
			(uint8_t*)dst, 720*2, 
			720, 576 );
	}
	else if (color_format==COLORFORMAT_RGB32)
	{
		YUV420_to_YUY2( (uint8_t*)dst_step1, 720, 
			(uint8_t*)(dst_step1+576*720/4+288*90), 360, 
			(uint8_t*)(dst_step1+576*720/4), 360,
			(uint8_t*)dst_step2, 720*2, 
			720, 576 );

		const int is_RGB2YUV_FULL = (pCtx->lpmyconfig->decoder_YUV_to_RGB_conversion == RGB2YUV_601 ? 0 : 1);

		if (pCtx->flip_RGB)
			mmx_YUY2toRGB32( (uint8_t*)dst_step2+(576-1)*720*2, (uint8_t*)dst, (uint8_t*)dst_step2-720*2, -720*2, 720*2, is_RGB2YUV_FULL);
		else
			mmx_YUY2toRGB32( (uint8_t*)dst_step2, (uint8_t*)dst, (uint8_t*)dst_step2+720*576*2, 720*2, 720*2, is_RGB2YUV_FULL);
	}
	else if (color_format==COLORFORMAT_RGB24)
	{
		YUV420_to_YUY2( (uint8_t*)dst_step1, 720, 
			(uint8_t*)(dst_step1+576*720/4+288*90), 360, 
			(uint8_t*)(dst_step1+576*720/4), 360,
			(uint8_t*)dst_step2, 720*2, 
			720, 576 );

		const int is_RGB2YUV_FULL = (pCtx->lpmyconfig->decoder_YUV_to_RGB_conversion == RGB2YUV_601 ? 0 : 1);

		if (pCtx->flip_RGB)
			mmx_YUY2toRGB24( (uint8_t*)dst_step2+(576-1)*720*2, (uint8_t*)dst, (uint8_t*)dst_step2-720*2, -720*2, 720*2, is_RGB2YUV_FULL);
		else
			mmx_YUY2toRGB24( (uint8_t*)dst_step2, (uint8_t*)dst, (uint8_t*)dst_step2+720*576*2, 720*2, 720*2, is_RGB2YUV_FULL);
	}
}


static void decode_NTSC_DVSD( void *hnd, uint8_t *src_DV, uint32_t *dst )
{
	decoder_context_t *pCtx = (decoder_context_t*)hnd;

	int color_format = pCtx->color_format;

	const int col_SB[5] = { 2, 1, 3, 0, 4};
	const int row_SB[10*5] = { 
		2, 3, 4, 5, 6, 7, 8, 9, 0, 1,
		6, 7, 8, 9, 0, 1, 2, 3, 4, 5,
		8, 9, 0, 1, 2, 3, 4, 5, 6, 7,
		0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
		4, 5, 6, 7, 8, 9, 0, 1, 2, 3
	};

	int i, j, k, p;
	int16_t *dst_5_MBs = pCtx->buffer_5_MBs;
	uint32_t *dst_step1 = dst;
	uint32_t *dst_step2 = pCtx->buffer_frame;

	if (color_format==COLORFORMAT_YUY2)
	{
		dst_step1 = pCtx->buffer_frame;
	}

	for (i=0; i<10; i++)	// sequence
	{
		for (k=0; k<27; k++)	// MB-index in SB
		{			
			int block_indx = 5*k; 
			int DBN_1st = (7+block_indx+block_indx/15);	// first DIF_blocknumber
			uint8_t *src_DIF_videosegment = src_DV + i*(150*80) + DBN_1st*(80);

			decode_DIF_videosegment( src_DIF_videosegment, dst_5_MBs, (uint8_t)i, DBN_1st );
			
			for(j=0; j<5; j++)	// superblock row number: a,b,c,d,e
			{
				p = col_SB[j];

				// save it first as YV12 and then convert to other formats
				// todo: direct conversion to the right format
				uint32_t *src = (uint32_t*)( dst_5_MBs + j*6*64 );

				int col_MB = MB_pos_NTSC[p][2*k];
				int row_MB = MB_pos_NTSC[p][2*k+1] + row_SB[j*10+i]*6;

				uint32_t *dst_y = dst_step1 + col_MB*(4*8/4) + row_MB*(8*720/4);
				uint32_t *dst_v = dst_step1 + (720*480/4) + col_MB*(8/4) + row_MB*(8*180/4);
				uint32_t *dst_u = dst_step1 + (720*480/4 + 180*480/4) + col_MB*(8/4) + row_MB*(8*180/4);
				int dst_y_pitch = 720;
				int dst_v_pitch = 180;
				int dst_u_pitch = dst_v_pitch;

				if ( (p==4) && (k>23) )
				{
					deinterleave_MBborder_YV12_NTSC(src, dst_y, dst_y_pitch, 
						dst_u, dst_u_pitch, dst_v, dst_v_pitch );
				}
				else
				{
					deinterleave_MB_YV12_NTSC(src, dst_y, dst_y_pitch, 
						dst_u, dst_u_pitch, dst_v, dst_v_pitch );
				}			
			}
		}
	}


	if (color_format==COLORFORMAT_YV12)
	{
		if (pCtx->lpmyconfig->decoder_YV12_chroma_sampling!=YV12_CHROMASAMPLING_DV)
		{
			YUV411_to_YUY2( (uint8_t*)dst_step1, 720, 
				(uint8_t*)(dst_step1+480*720/4+480*180/4), 180, 
				(uint8_t*)(dst_step1+480*720/4), 180,
				(uint8_t*)dst_step2, 720*2, 
				720, 480 );

			if (pCtx->lpmyconfig->decoder_YV12_chroma_sampling==YV12_CHROMASAMPLING_MPEG2_I)
			{
				mmx_yuy2_i_to_yv12( (uint8_t*)dst_step2, 720*2, 720*2, 
					(uint8_t*)dst, (uint8_t*)dst+480*720+240*360, (uint8_t*)dst+480*720, 720, 360,
					480 );
			}
			else
			{
				mmx_yuy2_to_yv12( (uint8_t*)dst_step2, 720*2, 720*2, 
					(uint8_t*)dst, (uint8_t*)dst+480*720+240*360, (uint8_t*)dst+480*720, 720, 360,
					480 );
			}
		}
	}
	else if (color_format==COLORFORMAT_YUY2)
	{
		YUV411_to_YUY2( (uint8_t*)dst_step1, 720, 
			(uint8_t*)(dst_step1+480*720/4+480*180/4), 180, 
			(uint8_t*)(dst_step1+480*720/4), 180,
			(uint8_t*)dst, 720*2, 
			720, 480 );
	}
	else if (color_format==COLORFORMAT_RGB32)
	{
		YUV411_to_YUY2( (uint8_t*)dst_step1, 720, 
			(uint8_t*)(dst_step1+480*720/4+480*180/4), 180, 
			(uint8_t*)(dst_step1+480*720/4), 180,
			(uint8_t*)dst_step2, 720*2, 
			720, 480 );

		const int is_RGB2YUV_FULL = (pCtx->lpmyconfig->decoder_YUV_to_RGB_conversion == RGB2YUV_601 ? 0 : 1);

		if (pCtx->flip_RGB)
			mmx_YUY2toRGB32( (uint8_t*)dst_step2+(480-1)*720*2, (uint8_t*)dst, (uint8_t*)dst_step2-720*2, -720*2, 720*2, is_RGB2YUV_FULL);
		else
			mmx_YUY2toRGB32( (uint8_t*)dst_step2, (uint8_t*)dst, (uint8_t*)dst_step2+720*480*2, 720*2, 720*2, is_RGB2YUV_FULL);
	}
	else if (color_format==COLORFORMAT_RGB24)
	{
		YUV411_to_YUY2( (uint8_t*)dst_step1, 720, 
			(uint8_t*)(dst_step1+480*720/4+480*180/4), 180, 
			(uint8_t*)(dst_step1+480*720/4), 180,
			(uint8_t*)dst_step2, 720*2, 
			720, 480 );

		const int is_RGB2YUV_FULL = (pCtx->lpmyconfig->decoder_YUV_to_RGB_conversion == RGB2YUV_601 ? 0 : 1);

		if (pCtx->flip_RGB)
			mmx_YUY2toRGB24( (uint8_t*)dst_step2+(480-1)*720*2, (uint8_t*)dst, (uint8_t*)dst_step2-720*2, -720*2, 720*2, is_RGB2YUV_FULL);
		else
			mmx_YUY2toRGB24( (uint8_t*)dst_step2, (uint8_t*)dst, (uint8_t*)dst_step2+720*480*2, 720*2, 720*2, is_RGB2YUV_FULL);
	}
}


static void decode_DIF_videosegment( uint8_t *DIF_block, int16_t *dst_5_MBs, 
								 uint32_t DIF_sequence_number, uint32_t DIF_block_number )
{
	uint32_t DCT_mode[5*6];
	uint32_t quantization_number[5]; 
	uint32_t class_number[5*6]; 
	int j, l, n;
	int16_t *block = dst_5_MBs;

	int iresult;

	iresult = decode_videosegment( dst_5_MBs, quantization_number, DCT_mode, class_number, 
									   DIF_block, DIF_sequence_number, DIF_block_number );

	n = 0;
	for (j=0; j<5; j++)
	{
		uint32_t QNO = quantization_number[j];

		for (l=0; l<6; l++)
		{
			int c1c0 = class_number[n];
			int mo = DCT_mode[n];
			int q_code = quantization_code_table_extended[(c1c0<<4)+QNO];
			
			if (mo==0)
			{
				int16_t *q_step = (int16_t*)quantization_step_88_extended + q_code*64;

				multiply_clip_16bit( block, block, q_step );
				block[0] = block[0]*16;
				weighting_block( block, block, (int16_t*)W_88_inv_16bit );

				idct_88( block, block);

				int k;
				for(k=0;k<64;k++) 
				{
					block[k] += 128;
					if ( (uint16_t)(block[k]) > 255 )
					{
						if (block[k]<0) 
							block[k]=0;
						else 
							block[k]=255;
					}
				}
			}
			else
			{
				int16_t *q_step = (int16_t*)quantization_step_248_extended + q_code*64;

				multiply_clip_16bit( block, block, q_step );
				block[0] = block[0]*16;
				weighting_block( block, block, (int16_t*)W_248_inv_16bit );

				idct_248( block, block);

				int k;
				for(k=0;k<64;k++) 
				{
					block[k] += 128;
					if ( (uint16_t)(block[k]) > 255 )
					{
						if (block[k]<0) 
							block[k]=0;
						else 
							block[k]=255;
					}
				}
			}

			n++;
			block += 64;
		}
	}
}


static void deinterleave_MB_YV12_PAL(uint32_t *src, uint32_t *dst_YUV420_Y, int dst_YUV420_Y_pitch, 
						  uint32_t *dst_YUV420_U, int dst_YUV420_U_pitch, 
					      uint32_t *dst_YUV420_V, int dst_YUV420_V_pitch
					      )
{
	uint8_t *x = (uint8_t*)dst_YUV420_Y;
	int16_t *y = (int16_t*)src;
	int x_pitch = dst_YUV420_Y_pitch*4;
	int i,k;

	for (k=0; k<2; k++)
	{
		for( i=0; i<8; i++)
		{
			// Block_Y0
			x[0] = (uint8_t)y[i*8+0];
			x[1] = (uint8_t)y[i*8+1];
			x[2] = (uint8_t)y[i*8+2];
			x[3] = (uint8_t)y[i*8+3];
			x[4] = (uint8_t)y[i*8+4];
			x[5] = (uint8_t)y[i*8+5];
			x[6] = (uint8_t)y[i*8+6];
			x[7] = (uint8_t)y[i*8+7];

			// Block_Y1
			x[0+8] = (uint8_t)y[i*8+64+0];
			x[1+8] = (uint8_t)y[i*8+64+1];
			x[2+8] = (uint8_t)y[i*8+64+2];
			x[3+8] = (uint8_t)y[i*8+64+3];
			x[4+8] = (uint8_t)y[i*8+64+4];
			x[5+8] = (uint8_t)y[i*8+64+5];
			x[6+8] = (uint8_t)y[i*8+64+6];
			x[7+8] = (uint8_t)y[i*8+64+7];

			x=x+x_pitch;
		}

		y+=128;
	}

	x = (uint8_t*)dst_YUV420_V;
	x_pitch = dst_YUV420_U_pitch*4;
	for( i=0; i<8; i++)
	{
		// Block_V
		x[0] = (uint8_t)y[i*8+0];
		x[1] = (uint8_t)y[i*8+1];
		x[2] = (uint8_t)y[i*8+2];
		x[3] = (uint8_t)y[i*8+3];
		x[4] = (uint8_t)y[i*8+4];
		x[5] = (uint8_t)y[i*8+5];
		x[6] = (uint8_t)y[i*8+6];
		x[7] = (uint8_t)y[i*8+7];

		x=x+x_pitch;
	}

	y+=64;

	x = (uint8_t*)dst_YUV420_U;
	x_pitch = dst_YUV420_V_pitch*4;
	for( i=0; i<8; i++)
	{
		// Block_V
		x[0] = (uint8_t)y[i*8+0];
		x[1] = (uint8_t)y[i*8+1];
		x[2] = (uint8_t)y[i*8+2];
		x[3] = (uint8_t)y[i*8+3];
		x[4] = (uint8_t)y[i*8+4];
		x[5] = (uint8_t)y[i*8+5];
		x[6] = (uint8_t)y[i*8+6];
		x[7] = (uint8_t)y[i*8+7];

		x=x+x_pitch;
	}
}


static void YUV420_to_YUY2( uint8_t* src_YUV420_Y, int src_YUV420_Y_pitch, 
					 uint8_t* src_YUV420_U, int src_YUV420_U_pitch, 
					 uint8_t* src_YUV420_V, int src_YUV420_V_pitch,
					 uint8_t* dst_YUY2, int dst_YUY2_pitch, int width, int height )
{
	int width_half = width/2;
	int height_minus_2 = height - 2;
	int x, y;
	int V_pitch_x2 = src_YUV420_V_pitch*2;
	int U_pitch_x2 = src_YUV420_U_pitch*2;

	uint8_t* out = dst_YUY2;
	uint8_t* Y = src_YUV420_Y;
	uint8_t* U = src_YUV420_U;
	uint8_t* V = src_YUV420_V;


	for( x=0; x<width_half; x++ )
	{
		out[0] = Y[0];
		out[1] = U[0];
		out[2] = Y[1];
		out[3] = V[0];
		out+=4;
		Y+=2;
		U++;
		V++;
	}
	out = out - 4*width_half + dst_YUY2_pitch;
	Y = Y - 2*width_half + src_YUV420_Y_pitch;
	U = U - width_half + src_YUV420_U_pitch;
	V = V - width_half + src_YUV420_V_pitch;

	for( x=0; x<width_half; x++ )
	{
		out[0] = Y[0];
		out[1] = U[0];
		out[2] = Y[1];
		out[3] = V[0];
		out+=4;
		Y+=2;
		U++;
		V++;
	}
	out = out - 4*width_half + dst_YUY2_pitch;
	Y = Y - 2*width_half + src_YUV420_Y_pitch;
	U = U - width_half - src_YUV420_U_pitch;
	V = V - width_half - src_YUV420_V_pitch;


	for ( y=2; y<height_minus_2; y+=4 )
	{

		for( x=0; x<width_half; x++ )
		{
			out[0] = Y[0];
			out[1] = U[0];
			out[2] = Y[1];
			out[3] = (uint8_t)( ((int)V[0]+(int)V[V_pitch_x2]+1)>>1 );
			out+=4;
			Y+=2;
			U++;
			V++;
		}
		out = out - 4*width_half + dst_YUY2_pitch;
		Y = Y - 2*width_half + src_YUV420_Y_pitch;
		U = U - width_half + src_YUV420_U_pitch;
		V = V - width_half + src_YUV420_V_pitch;

		for( x=0; x<width_half; x++ )
		{
			out[0] = Y[0];
			out[1] = U[0];
			out[2] = Y[1];
			out[3] = (uint8_t)( ((int)V[0]+(int)V[V_pitch_x2]+1)>>1 );
			out+=4;
			Y+=2;
			U++;
			V++;
		}
		out = out - 4*width_half + dst_YUY2_pitch;
		Y = Y - 2*width_half + src_YUV420_Y_pitch;
		U = U - width_half + src_YUV420_U_pitch;
		V = V - width_half + src_YUV420_V_pitch;

		for( x=0; x<width_half; x++ )
		{
			out[0] = Y[0];
			out[1] = (uint8_t)( ((int)U[-U_pitch_x2]+(int)U[0]+1)>>1 );
			out[2] = Y[1];
			out[3] = V[0];
			out+=4;
			Y+=2;
			U++;
			V++;
		}
		out = out - 4*width_half + dst_YUY2_pitch;
		Y = Y - 2*width_half + src_YUV420_Y_pitch;
		U = U - width_half + src_YUV420_U_pitch;
		V = V - width_half + src_YUV420_V_pitch;

		for( x=0; x<width_half; x++ )
		{
			out[0] = Y[0];
			out[1] = (uint8_t)( ((int)U[-U_pitch_x2]+(int)U[0]+1)>>1 );
			out[2] = Y[1];
			out[3] = V[0];
			out+=4;
			Y+=2;
			U++;
			V++;
		}
		out = out - 4*width_half + dst_YUY2_pitch;
		Y = Y - 2*width_half + src_YUV420_Y_pitch;
		U = U - width_half - src_YUV420_U_pitch;
		V = V - width_half - src_YUV420_V_pitch;
	}


	for( x=0; x<width_half; x++ )
	{
		out[0] = Y[0];
		out[1] = U[0];
		out[2] = Y[1];
		out[3] = V[0];
		out+=4;
		Y+=2;
		U++;
		V++;
	}
	out = out - 4*width_half + dst_YUY2_pitch;
	Y = Y - 2*width_half + src_YUV420_Y_pitch;
	U = U - width_half + src_YUV420_U_pitch;
	V = V - width_half + src_YUV420_V_pitch;

	for( x=0; x<width_half; x++ )
	{
		out[0] = Y[0];
		out[1] = U[0];
		out[2] = Y[1];
		out[3] = V[0];
		out+=4;
		Y+=2;
		U++;
		V++;
	}

}


static void deinterleave_MB_YV12_NTSC(uint32_t *src, uint32_t *dst_YUV420_Y, int dst_YUV420_Y_pitch, 
						  uint32_t *dst_YUV420_U, int dst_YUV420_U_pitch, 
					      uint32_t *dst_YUV420_V, int dst_YUV420_V_pitch
					      )
{
	int n, row;
	uint8_t *dst_y = (uint8_t *)dst_YUV420_Y;
	uint8_t *dst_u = (uint8_t *)dst_YUV420_U;
	uint8_t *dst_v = (uint8_t *)dst_YUV420_V;
	int stride_y = dst_YUV420_Y_pitch;
	int stride_u = dst_YUV420_U_pitch;
	int stride_v = dst_YUV420_V_pitch;
	int16_t *src_MB = (int16_t *)src;

	for (n=0; n<4; n++)
	{
		for ( row=0; row<8; row++)
		{
			dst_y[0] = (uint8_t)src_MB[0];
			dst_y[1] = (uint8_t)src_MB[1];
			dst_y[2] = (uint8_t)src_MB[2];
			dst_y[3] = (uint8_t)src_MB[3];
			dst_y[4] = (uint8_t)src_MB[4];
			dst_y[5] = (uint8_t)src_MB[5];
			dst_y[6] = (uint8_t)src_MB[6];
			dst_y[7] = (uint8_t)src_MB[7];
		
			src_MB += 8;
			dst_y += stride_y;
		}

		dst_y = dst_y - stride_y*8 + 8;
	}

	for ( row=0; row<8; row++)
	{
		dst_v[0] = (uint8_t)src_MB[0];
		dst_v[1] = (uint8_t)src_MB[1];
		dst_v[2] = (uint8_t)src_MB[2];
		dst_v[3] = (uint8_t)src_MB[3];
		dst_v[4] = (uint8_t)src_MB[4];
		dst_v[5] = (uint8_t)src_MB[5];
		dst_v[6] = (uint8_t)src_MB[6];
		dst_v[7] = (uint8_t)src_MB[7];
	
		src_MB += 8;
		dst_v += stride_v;
	}

	for ( row=0; row<8; row++)
	{
		dst_u[0] = (uint8_t)src_MB[0];
		dst_u[1] = (uint8_t)src_MB[1];
		dst_u[2] = (uint8_t)src_MB[2];
		dst_u[3] = (uint8_t)src_MB[3];
		dst_u[4] = (uint8_t)src_MB[4];
		dst_u[5] = (uint8_t)src_MB[5];
		dst_u[6] = (uint8_t)src_MB[6];
		dst_u[7] = (uint8_t)src_MB[7];
	
		src_MB += 8;
		dst_u += stride_u;
	}
}


static void deinterleave_MBborder_YV12_NTSC(uint32_t *src, uint32_t *dst_YUV420_Y, int dst_YUV420_Y_pitch, 
						  uint32_t *dst_YUV420_U, int dst_YUV420_U_pitch, 
					      uint32_t *dst_YUV420_V, int dst_YUV420_V_pitch
					      )
{
	int n, row, col;
	uint8_t *dst_y = (uint8_t *)dst_YUV420_Y;
	uint8_t *dst_u = (uint8_t *)dst_YUV420_U;
	uint8_t *dst_v = (uint8_t *)dst_YUV420_V;
	int stride_y = dst_YUV420_Y_pitch;
	int stride_u = dst_YUV420_U_pitch;
	int stride_v = dst_YUV420_V_pitch;
	int16_t *src_MB = (int16_t *)src;

	for (n=0; n<4; n++)
	{
		for ( row=0; row<8; row++)
		{
			for(col=0; col<8; col++)
				dst_y[col] = (uint8_t)src_MB[col];
		
			src_MB += 8;
			dst_y += stride_y;
		}

		if (n==1) 
			dst_y = dst_y - 8;
		else
			dst_y = dst_y - stride_y*8 + 8;

	}

	for ( row=0; row<8; row++)
	{
		for(col=0; col<4; col++)
			dst_v[col] = (uint8_t)src_MB[col];
	
		for(col=0; col<4; col++)
			dst_v[col+8*stride_v] = (uint8_t)src_MB[col+4];

		src_MB += 8;
		dst_v += stride_v;
	}

	for ( row=0; row<8; row++)
	{
		for(col=0; col<4; col++)
			dst_u[col] = (uint8_t)src_MB[col];
	
		for(col=0; col<4; col++)
			dst_u[col+8*stride_u] = (uint8_t)src_MB[col+4];

		src_MB += 8;
		dst_u += stride_u;
	}
}


static void YUV411_to_YUY2( uint8_t* src_YUV420_Y, int src_YUV420_Y_pitch, 
					 uint8_t* src_YUV420_U, int src_YUV420_U_pitch, 
					 uint8_t* src_YUV420_V, int src_YUV420_V_pitch,
					 uint8_t* dst_YUY2, int dst_YUY2_pitch, int width, int height )
{
	int row, col;
	int width_quarter = width/4;
	uint8_t *src_y = src_YUV420_Y;
	uint8_t *src_u = src_YUV420_U;
	uint8_t *src_v = src_YUV420_V;
	uint8_t *dst = dst_YUY2;

	for ( row=0; row<height; row++)
	{
		for( col=0; col<(width_quarter-1); col++ )
		{
			dst[col*8+0] = src_y[col*4+0];
			dst[col*8+1] = src_u[col  +0];
			dst[col*8+2] = src_y[col*4+1];
			dst[col*8+3] = src_v[col  +0];
			dst[col*8+4] = src_y[col*4+2];
			dst[col*8+5] = (uint8_t)(((int)src_u[col  +0] + (int)src_u[col  +1] + 1)>>1);
			dst[col*8+6] = src_y[col*4+3];
			dst[col*8+7] = (uint8_t)(((int)src_v[col  +0] + (int)src_v[col  +1] + 1)>>1);
		}
		dst[col*8+0] = src_y[col*4+0];
		dst[col*8+1] = src_u[col  +0];
		dst[col*8+2] = src_y[col*4+1];
		dst[col*8+3] = src_v[col  +0];
		dst[col*8+4] = src_y[col*4+2];
		dst[col*8+5] = src_u[col  +0];
		dst[col*8+6] = src_y[col*4+3];
		dst[col*8+7] = src_v[col  +0];

		dst += dst_YUY2_pitch;

		src_y += src_YUV420_Y_pitch;
		src_u += src_YUV420_U_pitch;
		src_v += src_YUV420_V_pitch;
	}
}

