/******************************************************************************/
/*                                                                            */
/*  Cedocida DV-Codec                                                         */
/*                                                                            */
/*  Copyright (C) 2005 Andreas Dittrich <dittrich@eit.uni-kl.de>              */
/*  Project start: 15.12.2003                                                 */
/*  Last change  : 15.06.2007                                                 */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  This program is free software; you can redistribute it and/or modify      */
/*  it under the terms of the GNU General Public License as published by      */
/*  the Free Software Foundation; either version 2 of the License, or         */
/*  (at your option) any later version.                                       */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*                                                                            */
/******************************************************************************/

#include <stdlib.h>
#include <memory.h>

#include "inttypes.h"

#include "dv_spec.h"
#include "dct/dct.h"
#include "vlc/encode_vlc.h"
#include "signalproc.h"
#include "encode_tables.h"
#include "config.h"
#include "encode.h"

// todo: do direct inplace conversion 
#include "convert/convert.h"


#define VIDEOSYSTEM_NTSC	0
#define VIDEOSYSTEM_PAL		1


typedef struct {
	uint8_t *buffer_base;

	int16_t *buffer_5_MBs;
	uint32_t *buffer_10_TMBs;
	uint32_t *buffer_frame;
	int16_t *block_tmp;

	int color_format;
	int flip_RGB;
	int is_PAL;

	configvars_t lpmyconfig[1];
} encoder_context_t;



static void encode_PAL_DVSD( void *hnd, uint32_t *src, uint32_t *dst_DV );
static void encode_NTSC_DVSD( void *hnd, uint32_t *src, uint32_t *dst_DV );

static void encode_6_Header_DIF_blocks( uint32_t *DIF_block, uint32_t DIF_sequence_number, char video_system, int is_16to9 );
static void encode_Audio_DIF_block( uint32_t *DIF_block, uint32_t DIF_sequence_number, uint32_t DIF_block_number, char video_system );
static void encode_5_Video_DIF_blocks( void *hnd, uint32_t *src_5_MBs, uint32_t *dst_5_DIF_blocks, uint8_t Dseq, uint8_t DBN_1st );
static void encode_videosegment( uint32_t *src_TMB, uint8_t *QNO, uint8_t *m0_c1_c0, uint32_t *dst_5_DIF_blocksint, uint8_t Dseq, uint8_t DBN_1st );




void * encoder_open()
{
	encoder_context_t *pCtx;

	pCtx = (encoder_context_t*)malloc( sizeof(encoder_context_t) );
	if (pCtx==NULL)
		return NULL;

#define LOG2_ALIGN 4
#define SIZE_BUFFER_5_MBS (5*6*64*2)
#define SIZE_BUFFER_10_TMBS (10*6*64*2)
#define SIZE_BUFFER_FRAME (720*576*4)
#define SIZE_BLOCK_TMP (64*2)

	pCtx->buffer_base = (uint8_t*)malloc(  ((1<<LOG2_ALIGN)-1) + 
		SIZE_BUFFER_FRAME + 
		SIZE_BUFFER_10_TMBS +
		SIZE_BUFFER_5_MBS +
		SIZE_BLOCK_TMP 	);
	if (pCtx->buffer_base==NULL)
		encoder_close( pCtx );

	pCtx->buffer_frame = (uint32_t*)(  ( (int64_t)(pCtx->buffer_base) + ((1<<LOG2_ALIGN)-1) ) & (~((1<<LOG2_ALIGN)-1))  );
	pCtx->buffer_10_TMBs = (uint32_t*)( (uint8_t*)(pCtx->buffer_frame) + SIZE_BUFFER_FRAME );
	pCtx->buffer_5_MBs = (int16_t*)( (uint8_t*)(pCtx->buffer_10_TMBs) + SIZE_BUFFER_10_TMBS );
	pCtx->block_tmp = (int16_t*)( (uint8_t*)(pCtx->buffer_5_MBs) + SIZE_BUFFER_5_MBS );

	pCtx->color_format = COLORFORMAT_YV12;
	pCtx->flip_RGB = 0;
	pCtx->is_PAL = 0;

	return pCtx;
}


void encoder_configure( void *hnd, int color_format, int flip_RGB, int is_PAL, configvars_t *lpmyconfig )
{
	encoder_context_t *pCtx = (encoder_context_t*)hnd;

	if (pCtx==NULL)
		return;

	pCtx->color_format = color_format;
	pCtx->flip_RGB = flip_RGB;
	pCtx->is_PAL = is_PAL;

	memcpy( pCtx->lpmyconfig, lpmyconfig, sizeof(configvars_t) );
}


void encoder_close( void *hnd )
{
	encoder_context_t *pCtx = (encoder_context_t*)hnd;

	if (pCtx==NULL)
		return;

	if (pCtx->buffer_base!=NULL)
		free(pCtx->buffer_base);

	free(pCtx);
	return;
}


void encode_DVSD( void *hnd, uint32_t *src, uint32_t *dst_DV )
{
	encoder_context_t *pCtx = (encoder_context_t*)hnd;

	if (pCtx==NULL)
		return;

	if (pCtx->is_PAL)
		encode_PAL_DVSD( pCtx, src, dst_DV );
	else
		encode_NTSC_DVSD( pCtx, src, dst_DV );

	emms_asm();
}


static void encode_PAL_DVSD( void *hnd, uint32_t *src, uint32_t *dst_DV )
{
	encoder_context_t *pCtx = (encoder_context_t*)hnd;

	int color_format = pCtx->color_format;

	//const int row_SB_offset[5] = { 2, 6, 8, 0, 4};
	static const int col_SB[5] = { 2, 1, 3, 0, 4};
	static const int row_SB[12*5] = { 
		2, 3, 4, 5, 6, 7, 8, 9,10,11, 0, 1,
		6, 7, 8, 9,10,11, 0, 1, 2, 3, 4, 5,
		8, 9,10,11, 0, 1, 2, 3, 4, 5, 6, 7,
		0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,
		4, 5, 6, 7, 8, 9,10,11, 0, 1, 2, 3
	};

	int i, r, j;

	int16_t * const dst_MB = pCtx->buffer_5_MBs;
	int16_t * const dst_MB_a = dst_MB + 0*6*64;
	int16_t * const dst_MB_b = dst_MB + 1*6*64;
	int16_t * const dst_MB_c = dst_MB + 2*6*64;
	int16_t * const dst_MB_d = dst_MB + 3*6*64;
	int16_t * const dst_MB_e = dst_MB + 4*6*64;

	// quick and dirty color-conversion (should be better done in combination with later "block-interleaving")
	if (color_format==COLORFORMAT_YV12) 
	{
		if ( pCtx->lpmyconfig->encoder_YV12_chroma_sampling==YV12_CHROMASAMPLING_MPEG2_P )
		{
			mmx_yv12_to_yuy2( (uint8_t*)src, (uint8_t*)(src+576*720/4+288*90), (uint8_t*)(src+576*720/4), 
								720, 720, 360, (uint8_t*)(pCtx->buffer_frame), 720*2, 576);
			color_format = COLORFORMAT_YUY2;
			src = pCtx->buffer_frame;
		}
		else if ( pCtx->lpmyconfig->encoder_YV12_chroma_sampling==YV12_CHROMASAMPLING_MPEG2_I )
		{
			mmx_yv12_i_to_yuy2( (uint8_t*)src, (uint8_t*)(src+576*720/4+288*90), (uint8_t*)(src+576*720/4), 
								720, 720, 360, (uint8_t*)(pCtx->buffer_frame), 720*2, 576);
			color_format = COLORFORMAT_YUY2;
			src = pCtx->buffer_frame;
		}
	}
	else if (color_format==COLORFORMAT_RGB32)
	{
		if (pCtx->flip_RGB)
			mmx_ConvertRGB32toYUY2_ad( src+((576-1)*720), pCtx->buffer_frame, -720, 720/2, 720, 576, pCtx->lpmyconfig->encoder_RGB_to_YUV_conversion );
		else
			mmx_ConvertRGB32toYUY2_ad( src, pCtx->buffer_frame, 720, 720/2, 720, 576, pCtx->lpmyconfig->encoder_RGB_to_YUV_conversion );

		color_format = COLORFORMAT_YUY2;
		src = pCtx->buffer_frame;
	}
	else if (color_format==COLORFORMAT_RGB24)
	{
		if (pCtx->flip_RGB)
			mmx_ConvertRGB24toYUY2_ad( src+((576-1)*540), pCtx->buffer_frame, -540, 720/2, 720, 576, pCtx->lpmyconfig->encoder_RGB_to_YUV_conversion );
		else
			mmx_ConvertRGB24toYUY2_ad( src, pCtx->buffer_frame, 540, 720/2, 720, 576, pCtx->lpmyconfig->encoder_RGB_to_YUV_conversion );

		color_format = COLORFORMAT_YUY2;
		src = pCtx->buffer_frame;
	}
	else if (color_format!=COLORFORMAT_YUY2)
	{
		return; //should never happen!
	}


	for (i=0; i<12; i++)	// sequence
	{
		encode_6_Header_DIF_blocks( dst_DV, i, VIDEOSYSTEM_PAL, pCtx->lpmyconfig->encoder_aspect_ratio );

		for (r=0; r<3; r++)	// row_MB in SB
		{
			uint32_t *src_yuy2;
			uint8_t * src_y = (uint8_t*)src;
			uint8_t *src_u, *src_v;

			int delta_y_a, delta_y_b, delta_y_c, delta_y_d, delta_y_e;
			int delta_uv_a, delta_uv_b, delta_uv_c, delta_uv_d, delta_uv_e;

			const int is_yv12 = (color_format == COLORFORMAT_YV12);
 
			if (is_yv12)
			{
				src_y = (uint8_t*)src;
				delta_y_a = col_SB[0]*9*16 + (row_SB[0*12+i]*3+r)*(16*720);
				delta_y_b = col_SB[1]*9*16 + (row_SB[1*12+i]*3+r)*(16*720);
				delta_y_c = col_SB[2]*9*16 + (row_SB[2*12+i]*3+r)*(16*720);
				delta_y_d = col_SB[3]*9*16 + (row_SB[3*12+i]*3+r)*(16*720);
				delta_y_e = col_SB[4]*9*16 + (row_SB[4*12+i]*3+r)*(16*720);
				src_v = src_y + 720*576;
				src_u = src_v + 360*288;
				delta_uv_a = col_SB[0]*9*8 + (row_SB[0*12+i]*3+r)*(8*360);
				delta_uv_b = col_SB[1]*9*8 + (row_SB[1*12+i]*3+r)*(8*360);
				delta_uv_c = col_SB[2]*9*8 + (row_SB[2*12+i]*3+r)*(8*360);
				delta_uv_d = col_SB[3]*9*8 + (row_SB[3*12+i]*3+r)*(8*360);
				delta_uv_e = col_SB[4]*9*8 + (row_SB[4*12+i]*3+r)*(8*360);
			}
			else //YUY2
			{
				src_yuy2 = src;
				delta_y_a = col_SB[0]*9*(16/2) + (row_SB[0*12+i]*3+r)*(16*720/2);
				delta_y_b = col_SB[1]*9*(16/2) + (row_SB[1*12+i]*3+r)*(16*720/2);
				delta_y_c = col_SB[2]*9*(16/2) + (row_SB[2*12+i]*3+r)*(16*720/2);
				delta_y_d = col_SB[3]*9*(16/2) + (row_SB[3*12+i]*3+r)*(16*720/2);
				delta_y_e = col_SB[4]*9*(16/2) + (row_SB[4*12+i]*3+r)*(16*720/2);
			}

			for (j=0; j<9; j++)	// col_MB in SB
			{
				uint8_t DBN_video = 5*SB_row_order[r][j]; //DIF_blocknumber of video
				uint8_t DBN_audio = DBN_video/15;	//DIF_blocknumber of audio
				uint8_t block_index = 7 + DBN_video + DBN_audio;
				uint32_t *dst_5_DIF_blocks = dst_DV + ((uint32_t)block_index)*(80/4);

				if ( (DBN_video%15)==0 )
				{
					encode_Audio_DIF_block( dst_5_DIF_blocks-(80/4), i, DBN_audio, VIDEOSYSTEM_PAL );
				}

				if (is_yv12)
				{
					// todo: MMX version
					copy_YV12_to_MBYVU420_12bit( src_y + delta_y_a, src_u + delta_uv_a, src_v + delta_uv_a, dst_MB_a, 720, 360 );
					copy_YV12_to_MBYVU420_12bit( src_y + delta_y_b, src_u + delta_uv_b, src_v + delta_uv_b, dst_MB_b, 720, 360 );
					copy_YV12_to_MBYVU420_12bit( src_y + delta_y_c, src_u + delta_uv_c, src_v + delta_uv_c, dst_MB_c, 720, 360 );
					copy_YV12_to_MBYVU420_12bit( src_y + delta_y_d, src_u + delta_uv_d, src_v + delta_uv_d, dst_MB_d, 720, 360 );
					copy_YV12_to_MBYVU420_12bit( src_y + delta_y_e, src_u + delta_uv_e, src_v + delta_uv_e, dst_MB_e, 720, 360 );
					src_y += 16;
					src_u += 8;
					src_v += 8;
				}
				else
				{
					copy_YUY2_to_MBYVU420_12bit( src_yuy2 + delta_y_a, dst_MB_a, 720*2 );
					copy_YUY2_to_MBYVU420_12bit( src_yuy2 + delta_y_b, dst_MB_b, 720*2 );
					copy_YUY2_to_MBYVU420_12bit( src_yuy2 + delta_y_c, dst_MB_c, 720*2 );
					copy_YUY2_to_MBYVU420_12bit( src_yuy2 + delta_y_d, dst_MB_d, 720*2 );
					copy_YUY2_to_MBYVU420_12bit( src_yuy2 + delta_y_e, dst_MB_e, 720*2 );
					src_yuy2 += 16/2;
				}

				encode_5_Video_DIF_blocks( pCtx, (uint32_t*)dst_MB, dst_5_DIF_blocks, (uint8_t)i, DBN_video );
			}
		}

		dst_DV += (150*80/4);
	}
}


static void encode_NTSC_DVSD( void *hnd, uint32_t *src, uint32_t *dst_DV )
{
	encoder_context_t *pCtx = (encoder_context_t*)hnd;

	int color_format = pCtx->color_format;

	static const int col_SB[5] = { 2, 1, 3, 0, 4};
	static const int row_SB[10*5] = { 
		2, 3, 4, 5, 6, 7, 8, 9, 0, 1,
		6, 7, 8, 9, 0, 1, 2, 3, 4, 5,
		8, 9, 0, 1, 2, 3, 4, 5, 6, 7,
		0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
		4, 5, 6, 7, 8, 9, 0, 1, 2, 3
	};

	int i, k, j, p;

	uint32_t *src_yuy2;
	uint8_t *src_y, *src_u, *src_v;

	int16_t *dst_MB = pCtx->buffer_5_MBs;

	// quick and dirty color-conversion (should be better done in combination with later "block-interleaving")
	if (color_format==COLORFORMAT_YV12) 
	{
		if ( pCtx->lpmyconfig->encoder_YV12_chroma_sampling==YV12_CHROMASAMPLING_MPEG2_P )
		{
			mmx_yv12_to_yuy2( (uint8_t*)src, (uint8_t*)(src+480*720/4+240*90), (uint8_t*)(src+480*720/4),
								720, 720, 360, (uint8_t*)(pCtx->buffer_frame), 720*2, 480);
			color_format = COLORFORMAT_YUY2;
			src = pCtx->buffer_frame;
		}
		else if ( pCtx->lpmyconfig->encoder_YV12_chroma_sampling==YV12_CHROMASAMPLING_MPEG2_I )
		{
			mmx_yv12_i_to_yuy2( (uint8_t*)src, (uint8_t*)(src+480*720/4+240*90), (uint8_t*)(src+480*720/4), 
								720, 720, 360, (uint8_t*)(pCtx->buffer_frame), 720*2, 480);
			color_format = COLORFORMAT_YUY2;
			src = pCtx->buffer_frame;
		}
	}
	else if (color_format==COLORFORMAT_RGB32)
	{
		if (pCtx->flip_RGB)
			mmx_ConvertRGB32toYUY2_ad( src+((480-1)*720), pCtx->buffer_frame, -720, 720/2, 720, 480, pCtx->lpmyconfig->encoder_RGB_to_YUV_conversion );
		else
			mmx_ConvertRGB32toYUY2_ad( src, pCtx->buffer_frame, 720, 720/2, 720, 480, pCtx->lpmyconfig->encoder_RGB_to_YUV_conversion );

		color_format = COLORFORMAT_YUY2;
		src = pCtx->buffer_frame;
	}
	else if (color_format==COLORFORMAT_RGB24)
	{
		if (pCtx->flip_RGB)
			mmx_ConvertRGB24toYUY2_ad( src+((480-1)*540), pCtx->buffer_frame, -540, 720/2, 720, 480, pCtx->lpmyconfig->encoder_RGB_to_YUV_conversion );
		else
			mmx_ConvertRGB24toYUY2_ad( src, pCtx->buffer_frame, 540, 720/2, 720, 480, pCtx->lpmyconfig->encoder_RGB_to_YUV_conversion );

		color_format = COLORFORMAT_YUY2;
		src = pCtx->buffer_frame;
	}
	else if (color_format!=COLORFORMAT_YUY2)
	{
		return; //should never happen!
	}


	for (i=0; i<10; i++)	// sequence
	{
		encode_6_Header_DIF_blocks( dst_DV, i, VIDEOSYSTEM_NTSC, pCtx->lpmyconfig->encoder_aspect_ratio );

		for (k=0; k<27; k++)	// MB-index in SB
		{
			uint8_t DBN_video = (uint8_t)( 5*k ); //DIF_blocknumber of video
			uint8_t DBN_audio = DBN_video/(uint8_t)15;	//DIF_blocknumber of audio
			uint8_t block_index = 7 + DBN_video + DBN_audio;
			uint32_t *dst_5_DIF_blocks = dst_DV + ((uint32_t)block_index)*(80/4);

			if ( (DBN_video%15)==0 )
			{
				encode_Audio_DIF_block( dst_5_DIF_blocks-(80/4), i, DBN_audio, VIDEOSYSTEM_NTSC );
			}

			for(j=0; j<5; j++)	// superblock row number: a,b,c,d,e
			{
				p = col_SB[j];
				int16_t *dst_MB_j = dst_MB + j*6*64;
				int col_MB = MB_pos_NTSC[p][2*k];
				int row_MB = MB_pos_NTSC[p][2*k+1] + row_SB[j*10+i]*6;

				if (color_format==COLORFORMAT_YV12)	// YV12
				{
					src_y = ((uint8_t*)src) + col_MB*4*8 + row_MB*(8*720);
					src_v = ((uint8_t*)src) + 720*480 + col_MB*8 + row_MB*(8*180);
					src_u = ((uint8_t*)src) + 720*480 +  + 180*480 + col_MB*8 + row_MB*(8*180);

					if ( (p==4) && (k>23) )
						copy_YV12_to_MBYVU411border_12bit( src_y, src_u, src_v, dst_MB_j, 720, 180 );
					else
						copy_YV12_to_MBYVU411_12bit( src_y, src_u, src_v, dst_MB_j, 720, 180 );
				}
				else if (color_format==COLORFORMAT_YUY2) //YUY2
				{
					src_yuy2 = src + col_MB*(16) + row_MB*(16*720/4);

					if ( (p==4) && (k>23) )
						copy_YUY2_to_MBYVU411border_12bit( src_yuy2, dst_MB_j, 720*2 );
					else
						copy_YUY2_to_MBYVU411_12bit( src_yuy2, dst_MB_j, 720*2, (col_MB==0 ? 1 : 0) );
				}
			}

			encode_5_Video_DIF_blocks( pCtx, (uint32_t*)dst_MB, dst_5_DIF_blocks, (uint8_t)i, DBN_video );
		}

		dst_DV += (150*80/4);
	}
}


static void encode_6_Header_DIF_blocks( uint32_t *DIF_block, uint32_t DIF_sequence_number, char video_system, int is_16to9 )
{
	uint8_t* PC;	// packet pointer

	const uint32_t SCT_head = 0;	// DIF-block-type: head
	const uint32_t SCT_subcode = 1;	// DIF-block-type: subcode
	const uint32_t SCT_vaux = 2;	// DIF-block-type: vaux
	const uint32_t Seq = 15;	// video-sequencenumber: n/a
	uint32_t Dseq = DIF_sequence_number;
	
	uint8_t bcsys = 0;
	uint8_t disp = 0;

	if (is_16to9)
	{
		if (video_system == VIDEOSYSTEM_NTSC)
		{
			bcsys = 0;
			disp = 0x02;
		}
		else
		{
			bcsys = 1;
			disp = 0x07;
		}
	}
	
	// set all bytes to 0xFF
	ones_block80( DIF_block, 6 );	

	// head
	DIF_block[20*0] =  0xFF000000 | ((0)<<16) | (Dseq<<12) | 0x0700 | (SCT_head<<5) | 0x10 |  (Seq);	//byteswaped!

	PC = ((uint8_t*)(DIF_block)) + 3;	// head data
	PC[0] = 0x3F | (video_system<<7);
	PC[1] = 0xF8;	//  DFTIA=1111: n/a, 1, APT=100
	PC[2] = 0x78;	// 
	PC[3] = 0x78;	//  
	PC[4] = 0x78;	// 

	// subcode 0, 1
	DIF_block[20*1] =  0xFF000000 | ((0)<<16) | (Dseq<<12) | 0x0700 | (SCT_subcode<<5) | 0x10 | (Seq);	//byteswaped!
	DIF_block[20*2] =  0xFF000000 | ((1)<<16) | (Dseq<<12) | 0x0700 | (SCT_subcode<<5) | 0x10 | (Seq);	//byteswaped!

	// vaux 0, 1, 2
	DIF_block[20*3] =  0xFF000000 | ((0)<<16) | (Dseq<<12) | 0x0700 | (SCT_vaux<<5) | 0x10 |  (Seq);	//byteswaped!
	DIF_block[20*4] =  0xFF000000 | ((1)<<16) | (Dseq<<12) | 0x0700 | (SCT_vaux<<5) | 0x10 |  (Seq);	//byteswaped!
	DIF_block[20*5] =  0xFF000000 | ((2)<<16) | (Dseq<<12) | 0x0700 | (SCT_vaux<<5) | 0x10 |  (Seq);	//byteswaped!

	if ( DIF_sequence_number & 0x01 )	// even or odd? see: IEC 61834-2 / table 32
		PC = ((uint8_t*)(DIF_block+20*3)) + 3;			// video-packet 0
	else
		PC = ((uint8_t*)(DIF_block+20*5)) + 3 + 9*5;	// video-packet 39

	PC[0] = 0x60;	// source
	PC[1] = 0xFF;	// no information
	PC[2] = 0xFF;	// CLF not valid
	PC[3] = 0xC0 | (video_system<<5);	// system 625-50 (PAL)
	PC[4] = 0xFF;	// no information

	PC+=5;
	PC[0] = 0x61;	// source control
	PC[1] = 0x3F;	// CGMS=00: copy allowed, ISR=11: n/a, CMP=11: n/a, SS=11: n/a
	PC[2] = 0xC8 | disp;	// REC ST=1: no startpoint, 1, REC MODE=00: original, 1, disp
	PC[3] = 0xFC | bcsys;	// FF=1: both fields, FS=1: field1 first, FC=1: different frame, IL=1: interlaced, ST=1: no picture, SC=1: no picture, BCSYS
	PC[4] = 0xFF;	// 1, category=1111111

	// next optional packets are date, time ...

}


static void encode_Audio_DIF_block( uint32_t *DIF_block, uint32_t DIF_sequence_number, uint32_t DIF_block_number, char video_system )
{
	uint8_t* PC;	// packet pointer

	const uint32_t SCT = 3;	// DIF-block-type: audio
	const uint32_t Seq = 15;	// Video-sequenznumber: not available
	uint32_t Dseq = DIF_sequence_number;
	uint32_t DBN = DIF_block_number;

	// set all bytes to zero
	clear_block80( DIF_block, 1 );	

	*DIF_block = (DBN<<16) | (Dseq<<12) | 0x0700 | (SCT<<5) | 0x10 | (Seq);	//byteswaped!

	PC = ((uint8_t*)(DIF_block)) + 3;	// AAUX packet

	if ( (DIF_sequence_number & 0x01)==0 )	// even?	see: IEC 61834-2 / table 31
		DIF_block_number = DIF_block_number - 3;	// only to combine the next if-section

	if ( DIF_block_number==0 )
	{
		PC[0] = 0x50;	// source 
		PC[1] = 0x40;	// LF=0: unlocked mode, 1, AF-size=000000
		PC[2] = 0x0F;	// SM=0, CHN=00, PA=0, audio-mode=1111: n/a, 
		PC[3] = 0xC0 | (video_system<<5);	// 1, ML=1, system=x, stype=00000
		PC[4] = 0xC0;	// EF=1, TC=1, SMP=000: 48kHz, QU=000: 16bit
	}
	else if ( DIF_block_number==1 )
	{
		PC[0] = 0x51;	// source control
		PC[1] = 0x3F;	// CGMS=00: copy allowed, ISR=11: n/a, CMP=11: n/a, SS=11: n/a
		PC[2] = 0xFF;	// REC ST=1: no startpoint, REC END=1, REC MODE=111: not valid, Insert-CH=111: n/a
		PC[3] = 0xFF;	// DRF=1: forward, speed=1111111: n/a
		PC[4] = 0xFF;	// 1, category=1111111
	}
	else	// next 4 optional packets are e.g. date, ...
	{
		PC[0] = 0xFF;	// NOI-packet 
		PC[1] = 0xFF;	//  
		PC[2] = 0xFF;	//  
		PC[3] = 0xFF;	//  
		PC[4] = 0xFF;	// 
	}
}


static void encode_5_Video_DIF_blocks( void *hnd, uint32_t *src_5_MBs, uint32_t *DIF_block, uint8_t DIF_sequence_number, uint8_t DIF_block_number )
{
	encoder_context_t *pCtx = (encoder_context_t*)hnd;

	uint32_t DCT_mode[5*6];
	uint32_t quantization_number[5]; 
	uint32_t class_number[5*6]; 
	int dct248_threshold_x16 = pCtx->lpmyconfig->encoder_dct248_threshold*16;
	int dct248_counts = pCtx->lpmyconfig->encoder_dct248_counts;
	int length_test, decrease_QNO;
	int QNO;
	short *core;

	int i;
	int class_numbers_adjusted;

	// do DCT-based tranforming on all 5 MBs and do classification
	for (i=0; i<(5*6); i++)
	{
		int max_abs_value;
		int block_number = (i%6);

		int16_t *src_MB = (int16_t*)src_5_MBs + i*64;
		int16_t *block = (int16_t*)(pCtx->buffer_10_TMBs) + i*64;

		if (is_interlaced(src_MB, dct248_threshold_x16) < dct248_counts)
		{
			dct_88_weighted( src_MB, block);
			//block[0] = block[0] - 2048;
			DCT_mode[i] = 0;
		}
		else
		{
			dct_248_weighted( src_MB, block);
			//block[0] = block[0] - 2048;
			DCT_mode[i] = 1;
		}

		max_abs_value = max_abs_AC( block )/16;

		if (block_number<4)	// Y
		{
			if ( max_abs_value >= pCtx->lpmyconfig->encoder_classification_Y[2])
				class_number[i]=3;
			else if ( max_abs_value >= pCtx->lpmyconfig->encoder_classification_Y[1])
				class_number[i]=2;
			else if ( max_abs_value >= pCtx->lpmyconfig->encoder_classification_Y[0])
				class_number[i]=1;
			else
				class_number[i]=0;
		}
		else if (block_number==4)	// V or CR
		{
			if ( max_abs_value >= pCtx->lpmyconfig->encoder_classification_V[2])
				class_number[i]=3;
			else if ( max_abs_value >= pCtx->lpmyconfig->encoder_classification_V[1])
				class_number[i]=2;
			else if ( max_abs_value >= pCtx->lpmyconfig->encoder_classification_V[0])
				class_number[i]=1;
			else
				class_number[i]=0;
		}
		else	// U or CB
		{
			if ( max_abs_value >= pCtx->lpmyconfig->encoder_classification_U[2])
				class_number[i]=3;
			else if ( max_abs_value >= pCtx->lpmyconfig->encoder_classification_U[1])
				class_number[i]=2;
			else if ( max_abs_value >= pCtx->lpmyconfig->encoder_classification_U[0])
				class_number[i]=1;
			else
				class_number[i]=0;
		}
	}


	core = core_positive_negative;
	//apply user defined coring_c pattern if enabled in settings
	//coring( (int16_t*)buffer_10_TMBs, core, 5*6 );

#define LENGTH_TEST_MAX (5*67*8)

	switch ( pCtx->lpmyconfig->encoder_quality_mode )
	{

	case QUALITY_MODE_FAST:
		QNO=12;
		class_numbers_adjusted=0;
		do {
			length_test = 0;
			decrease_QNO = 0;
			for (i=0; i<(5*6); i++)
			{
				int16_t *block = (int16_t*)(pCtx->buffer_10_TMBs) + i*64;
				int c1c0 = class_number[i];
				int q_code = quantization_code_table_extended[(c1c0<<4)+QNO];
				int mo = DCT_mode[i];
				int length_block;

				if (mo==0)
				{
					int16_t *q_scale = (int16_t*)quantization_scale_88_extended + q_code*64;
					weighting_clipping( block, pCtx->block_tmp, q_scale );
					length_block = get_vlc_length_88_unrolled( pCtx->block_tmp );
				}
				else
				{
					int16_t *q_scale = (int16_t*)quantization_scale_248_extended + q_code*64;
					weighting_clipping( block, pCtx->block_tmp, q_scale );
					length_block = get_vlc_length_248_unrolled( pCtx->block_tmp );
				}

				length_test += length_block;

				if (length_test>LENGTH_TEST_MAX) 
				{
					decrease_QNO = 4;
					break;
				}
			}	
			
			QNO = QNO - decrease_QNO;

			if (QNO<0)
			{
				if ((QNO+decrease_QNO)!=0)
				{
					QNO = 0;
				}
				else if (!class_numbers_adjusted)
				{
					for (i=0; i<(5*6); i++)
						class_number[i] = 3;
					class_numbers_adjusted = 1;
					QNO = 8;
				}
				else if ( (core-core_positive_negative+128) < sizeof(core_positive_negative) )
				{
					QNO = 8;
					core = core+128;
					coring( (int16_t*)(pCtx->buffer_10_TMBs), core, 5*6 );
				}
				else	// should never happen!!!
				{
					QNO = -1;	
					decrease_QNO = 0;
				}
			}
			
		} while ( decrease_QNO>0 );
		
		break;

	default:	// QUALITY_MODE_NORMAL
		QNO=15;
		class_numbers_adjusted = 0;
		do {
			length_test = 0;
			decrease_QNO = 0;
			for (i=0; i<(5*6); i++)
			{
				int16_t *block = (int16_t*)(pCtx->buffer_10_TMBs) + i*64;
				int c1c0 = class_number[i];
				int q_code = quantization_code_table_extended[(c1c0<<4)+QNO];
				int mo = DCT_mode[i];
				int length_block;

				if (mo==0)
				{
					int16_t *q_scale = (int16_t*)quantization_scale_88_extended + q_code*64;
					weighting_clipping( block, pCtx->block_tmp, q_scale );
					length_block = get_vlc_length_88_unrolled( pCtx->block_tmp );
				}
				else
				{
					int16_t *q_scale = (int16_t*)quantization_scale_248_extended + q_code*64;
					weighting_clipping( block, pCtx->block_tmp, q_scale );
					length_block = get_vlc_length_248_unrolled( pCtx->block_tmp );
				}

				length_test += length_block;

				if (length_test>LENGTH_TEST_MAX) 
				{
					decrease_QNO = 1;
					break;
				}
			}	
			
			QNO = QNO - decrease_QNO;

			if (QNO<0)
			{
				if (!class_numbers_adjusted)
				{
					for (i=0; i<(5*6); i++)
						class_number[i] = 3;
					class_numbers_adjusted = 1;
					QNO = 15;
				}
				else if ( (core-core_positive_negative+128) < sizeof(core_positive_negative) )
				{
					QNO = 12;
					core = core+128;
					coring( (int16_t*)(pCtx->buffer_10_TMBs), core, 5*6 );
				}
				else	// should never happen!!!
				{
					QNO = -1;	
					decrease_QNO = 0;
				}
			}
			
		} while ( decrease_QNO>0 );
		
		break;

	}

	if (QNO<0) 
	{
		return;	// for debugging only
	}

	for (i=0; i<(5*6); i++)
	{
		int c1c0 = class_number[i];
		int q_code = quantization_code_table_extended[(c1c0<<4)+QNO];
		int mo = DCT_mode[i];
		int16_t *block = (int16_t*)(pCtx->buffer_10_TMBs) + i*64;
		
		if (mo==0)
		{
			int16_t *q_scale = (int16_t*)quantization_scale_88_extended + q_code*64;
			weighting_clipping( block, block, q_scale );
		}
		else
		{
			int16_t *q_scale = (int16_t*)quantization_scale_248_extended + q_code*64;
			weighting_clipping( block, block, q_scale );
		}
	}		


	for (i=0; i<(5); i++)
	{
		quantization_number[i] = QNO;
	}

	int y = encode_videosegment( (int16_t*)(pCtx->buffer_10_TMBs), quantization_number, DCT_mode, class_number, 
			(uint8_t*)DIF_block, DIF_sequence_number, DIF_block_number );

	if (y>0)
	{
		return;	// for debugging only
	}
}



